import logging
from time import sleep

from kazoo.exceptions import NodeExistsError

from bubuku.amazon import Amazon
from bubuku.broker import BrokerManager
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.controller')


class Change(object):
    def get_name(self) -> str:
        raise NotImplementedError('Not implemented yet')

    def can_run(self, current_actions) -> bool:
        raise NotImplementedError('Not implemented yet')

    def run(self, current_actions) -> bool:
        raise NotImplementedError('Not implemented')

    def can_run_at_exit(self) -> bool:
        return False


class Check(object):
    def check(self) -> Change:
        raise NotImplementedError('Not implemented')


def _exclude_self(ip, name, running_actions):
    return [k for k, v in running_actions.items() if k != name and v != ip]


class Controller(object):
    def __init__(self, broker_manager: BrokerManager, zk: Exhibitor, amazon: Amazon):
        self.broker_manager = broker_manager
        self.zk = zk
        self.amazon = amazon
        self.checks = []
        self.changes = []
        self.running = True

    def add_check(self, check):
        _LOG.info('Adding check {}'.format(str(check)))
        self.checks.append(check)

    def loop(self):
        ip = self.amazon.get_own_ip()

        try:
            self.zk.create('/bubuku/changes', makepath=True)
        except NodeExistsError:
            pass
        while self.running or self.changes:
            _LOG.debug('Taking lock for processing')
            with self.zk.take_lock('/bubuku/global_lock', ip):
                _LOG.debug('Lock is taken')
                # Get list of current running changes
                running_changes = {
                    change: self.zk.get('/bubuku/changes/{}'.format(change))[0].decode('utf-8')
                    for change in self.zk.get_children('/bubuku/changes')
                    }
                if running_changes:
                    _LOG.info("Running changes: {}".format(running_changes))
                # Register changes to run
                for change in self.changes:
                    name = change.get_name()
                    if change.can_run(_exclude_self(ip, name, running_changes)):
                        if name not in running_changes:
                            _LOG.info('Registering change in zk: {}'.format(name))
                            self.zk.create('/bubuku/changes/{}'.format(name), ip.encode('utf-8'), ephemeral=True)
                            running_changes[name] = ip
                    else:
                        _LOG.info('Change {} is waiting for others: {}'.format(change.get_name(), running_changes))
            # apply changes without holding lock
            actions_to_remove = []
            for change in self.changes:
                if change.get_name() in running_changes and running_changes[change.get_name()] == ip:
                    _LOG.info('Executing action {} step'.format(change.get_name()))
                    if self.running or change.can_run_at_exit():
                        if not change.run(_exclude_self(ip, change.get_name(), running_changes)):
                            _LOG.info('Action {} completed'.format(change.get_name()))
                            actions_to_remove.append(change.get_name())
                        else:
                            _LOG.info('Action {} will be executed on next loop step'.format(change.get_name()))
                    else:
                        _LOG.info(
                            'Action {} can not be run while stopping, forcing to stop it'.format(change.get_name()))
                        actions_to_remove.append(change.get_name())

            # remove processed actions
            if actions_to_remove:
                self.changes = [ch for ch in self.changes if ch.get_name() not in actions_to_remove]
                with self.zk.take_lock('/bubuku/global_lock'):
                    for name in actions_to_remove:
                        _LOG.info('Removing action {} from locks'.format(name))
                        self.zk.delete('/bubuku/changes/{}'.format(name), recursive=True)

            if self.running:
                for check in self.checks:
                    _LOG.info('Executing check {}'.format(check))
                    change = check.check()
                    if change:
                        _LOG.info('Adding change {} to pending changes'.format(change.get_name()))
                        self.changes.append(change)

            if self.changes:
                sleep(0.5)
            else:
                sleep(5)

    def stop(self, change: Change):
        _LOG.info('Stopping controller with additional change: {}'.format(change.get_name() if change else None))
        # clear all pending changes
        if change:
            self.changes.append(change)
        self.running = False



import json
import logging

from kazoo.exceptions import NodeExistsError, NoNodeError

from bubuku.broker import BrokerManager
from bubuku.controller import Check, Change
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.features.rebalance')


def _optimise_broker_ids(ids: list) -> str:
    if len(ids) > 1:
        ids[1:] = sorted(ids[1:])
    return ','.join(ids)


def combine_broker_ids(ids, length) -> list:
    result = []

    def _combine(start, left, length_):
        if not left or length_ == 0:
            result.append(start)
            return
        for i in range(0, len(left)):
            left_copy = list(left)
            del left_copy[i]
            next_copy = list(start)
            next_copy.append(left[i])
            _combine(next_copy, left_copy, length_ - 1)

    _combine([], ids, length)

    return sorted(set([_optimise_broker_ids(x) for x in result]))


def pop_with_length(arrays, length):
    for d in arrays:
        if len(d) == length:
            def _removal_func():
                del d[0]

            return d[0], _removal_func
    return None, None


class RebalanceChange(Change):
    def __init__(self, zk: Exhibitor, broker_list):
        self.zk = zk
        self.broker_ids = broker_list
        self.stale_data = {}  # partition count to topic data
        self.shuffled_broker_ids = None

    def get_name(self) -> str:
        return 'rebalance'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['start', 'restart', 'rebalance', 'stop']])

    def load_current_data(self):
        result = []
        for topic in self.zk.get_children('/brokers/topics'):
            data = json.loads(self.zk.get("/brokers/topics/" + topic)[0].decode('utf-8'))
            for k, v in data['partitions'].items():
                result.append({"topic": topic, "partition": int(k), "replicas": v})
        return result

    def take_next(self) -> dict:
        if self.stale_data:
            k = [x for x in self.stale_data.keys()][0]
            v = self.stale_data[k]

            def _removal_func():
                del v[0]
                if not v:
                    del self.stale_data[k]

            return v[0], _removal_func

        if self.shuffled_broker_ids:
            for repl_factor, data in self.shuffled_broker_ids.items():
                min_length = min([len(v) for v in data.values()])
                max_length = max([len(v) for v in data.values()])
                if (max_length - min_length) < 2:
                    continue
                return pop_with_length(data.values(), max_length)
        return None, None

    def run(self, current_actions):
        # Stop rebalance if someone is restarting
        if any([a in current_actions for a in ['restart', 'start', 'stop']]):
            _LOG.warning("Rebalance stopped, because other blocking events running: {}".format(current_actions))
            return False

        try:
            rebalance_data = self.zk.get('/admin/reassign_partitions')[0].decode('utf-8')
            _LOG.info('Old rebalance is still in progress: {}, waiting'.format(rebalance_data))
            return True
        except NoNodeError:
            pass

        new_broker_ids = sorted(self.zk.get_children('/brokers/ids'))

        if new_broker_ids != self.broker_ids:
            _LOG.warning("Rebalance stopped because of broker list change from {} to {}".format(self.broker_ids,
                                                                                                new_broker_ids))
            return False

        # Next actions are split into steps, because they are relatively long-running

        # Load existing data from zookeeper and try to split it for different purposes
        if self.shuffled_broker_ids is None:
            self.shuffled_broker_ids = {}
            for d in self.load_current_data():
                replication_factor = len(d['replicas'])
                if replication_factor > len(self.broker_ids):
                    _LOG.warning(
                        "Will not rebalance partition {} because only {} brokers available".format(d, self.broker_ids))
                    continue
                if replication_factor not in self.shuffled_broker_ids:
                    self.shuffled_broker_ids[replication_factor] = {
                        k: [] for k in combine_broker_ids(self.broker_ids, replication_factor)
                        }
                name = _optimise_broker_ids([str(i) for i in d['replicas']])
                if name not in self.shuffled_broker_ids[replication_factor]:
                    if name not in self.stale_data:
                        self.stale_data[name] = []
                    self.stale_data[name].append(d)
                else:
                    self.shuffled_broker_ids[replication_factor][name].append(d)
            _LOG.info("Shuffled broker ids are: {}".format(json.dumps(self.shuffled_broker_ids, indent=2)))
            return True

        to_move, removal_func = self.take_next()
        while to_move:
            replication_factor = len(to_move['replicas'])
            if replication_factor not in self.shuffled_broker_ids:
                removal_func()
                _LOG.error("Well, it's a BUG! Replication factor {} is not found among {}".format(
                    replication_factor,
                    self.shuffled_broker_ids.keys()))
            else:
                min_length = min([len(v) for v in self.shuffled_broker_ids[replication_factor].values()])
                for k, v in self.shuffled_broker_ids[replication_factor].items():
                    if len(v) == min_length:
                        j = {
                            "version": "1",
                            "partitions": [
                                {
                                    "topic": to_move['topic'],
                                    "partition": to_move['partition'],
                                    "replicas": [int(v) for v in k.split(',')],
                                }
                            ]
                        }
                        if self.reallocate(j):
                            _LOG.info("Current allocation: \n{}".format(self.dump_allocations()))
                            removal_func()
                            v.append(to_move)
                        return True
            to_move, removal_func = self.take_next()

        _LOG.info("Current allocation: \n{}".format(self.dump_allocations()))
        return False

    def reallocate(self, j: dict):
        try:
            data = json.dumps(j)
            self.zk.create("/admin/reassign_partitions", data.encode('utf-8'))
            _LOG.info("Reallocating {}".format(data))
            return True
        except NodeExistsError:
            _LOG.info("Waiting for free reallocation slot, still in progress...")
            return False

    def dump_allocations(self):
        return '\n'.join(
            ['\n'.join(['{}:{}'.format(x, len(y)) for x, y in v.items()]) for v in self.shuffled_broker_ids.values()])


class RebalanceOnStartCheck(Check):
    def __init__(self, zk, broker: BrokerManager):
        self.zk = zk
        self.broker = broker
        self.executed = False

    def check(self):
        if self.executed:
            return None
        if not self.broker.is_running_and_registered():
            return None
        _LOG.info("Rebalance on start, triggering rebalance")
        self.executed = True
        return RebalanceChange(self.zk, sorted(self.zk.get_children('/brokers/ids')))


class RebalanceOnBrokerListChange(Check):
    def __init__(self, zk, broker: BrokerManager):
        self.zk = zk
        self.broker = broker
        self.old_broker_list = []

    def check(self):
        if not self.broker.is_running_and_registered():
            return None
        new_list = sorted(self.zk.get_children('/brokers/ids'))
        if not new_list == self.old_broker_list:
            _LOG.info('Broker list changed from {} to {}, triggering rebalance'.format(self.old_broker_list, new_list))
            self.old_broker_list = new_list
            return RebalanceChange(self.zk, new_list)
        return None



import logging
import os
from collections import namedtuple

_LOG = logging.getLogger('bubuku.properties')

Config = namedtuple('Config', ('kafka_dir', 'kafka_settings_template', 'zk_stack_name',
                               'zk_prefix', 'id_policy', 'features', 'health_port'))


class KafkaProperties(object):
    def __init__(self, template: str, kafka_settings: str):
        self.lines = []
        self.settings_file = kafka_settings
        _LOG.info('Loading template properties from {}'.format(self.settings_file))
        with open(template, 'r') as f:
            for l in f.readlines():
                self.lines.append(_make_clean_line(l))

    def get_property(self, name: str) -> str:
        idx = self._get_property_idx(name)
        if idx is not None:
            return self.lines[idx].split('=', 1)[1]
        return None

    def _get_property_idx(self, name: str):
        search = '{}='.format(name)
        for idx in range(0, len(self.lines)):
            if self.lines[idx].startswith(search):
                return idx
        return None

    def delete_property(self, name):
        idx = self._get_property_idx(name)
        if idx is not None:
            del self.lines[idx]

    def set_property(self, name, value):
        idx = self._get_property_idx(name)
        if idx is not None:
            self.lines[idx] = '{}={}'.format(name, value)
        else:
            self.lines.append('{}={}'.format(name, value))

    def dump(self):
        _LOG.info('Dumping kafka properties to {}'.format(self.settings_file))
        with open(self.settings_file, mode='w') as f:
            for l in self.lines:
                f.write('{}\n'.format(l))


def load_config() -> Config:
    zk_prefix = os.getenv('ZOOKEEPER_PREFIX', '/')
    return Config(
        kafka_dir=os.getenv('KAFKA_DIR'),
        kafka_settings_template=os.getenv('KAFKA_SETTINGS'),
        zk_stack_name=os.getenv('ZOOKEEPER_STACK_NAME'),
        zk_prefix=zk_prefix if zk_prefix.startswith('/') or not zk_prefix else '/{}'.format(zk_prefix),
        id_policy=os.getenv('BROKER_ID_POLICY', 'ip').lower(),
        features=os.getenv('BUKU_FEATURES', '').lower(),
        health_port=int(os.getenv('HEALTH_PORT', '8888')),
    )


def _make_clean_line(l: str) -> str:
    result = l.strip()
    if result.startswith('#') or not result:
        return result
    n, v = result.split('=', 1)
    return '{}={}'.format(n.strip(), v)



import json
import logging
import subprocess

from bubuku.config import KafkaProperties
from bubuku.id_generator import BrokerIdGenerator
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.broker')


class LeaderElectionInProgress(Exception):
    pass


class BrokerManager(object):
    def __init__(self, kafka_dir: str, exhibitor: Exhibitor, id_manager: BrokerIdGenerator,
                 kafka_properties: KafkaProperties):
        self.kafka_dir = kafka_dir
        self.id_manager = id_manager
        self.exhibitor = exhibitor
        self.kafka_properties = kafka_properties
        self.process = None
        self.wait_timeout = 5 * 60

    def is_running_and_registered(self):
        if not self.process:
            return False
        return self.id_manager.is_registered()

    def stop_kafka_process(self):
        self._terminate_process()
        self._wait_for_zk_absence()

    def _is_clean_election(self):
        value = self.kafka_properties.get_property('unclean.leader.election.enable')
        return value == 'false'

    def has_leadership(self):
        """
        Says if this broker is still a leader for partitions or in ISR list for some partitions
        :return: True, if broker is a leader or have isr.
        """
        # Only wait when unclean leader election is disabled
        if not self._is_clean_election():
            return False
        broker_id = str(self.id_manager.get_broker_id())
        if not broker_id:
            return False
        for topic in self.exhibitor.get_children('/brokers/topics'):
            for partition in self.exhibitor.get_children('/brokers/topics/{}/partitions'.format(topic)):
                state = json.loads(
                    self.exhibitor.get('/brokers/topics/{}/partitions/{}/state'.format(topic, partition))[0].decode(
                        'utf-8'))
                if str(state['leader']) == broker_id:
                    _LOG.warn('Broker {} is still a leader for {} {} ({})'.format(broker_id, topic, partition,
                                                                                  json.dumps(state)))
                    return True
                if any([str(x) == broker_id for x in state['isr']]):
                    _LOG.warn('Broker {} is still is in ISR for {} {} ({})'.format(broker_id, topic, partition,
                                                                                   json.dumps(state)))
                    return True
        return False

    def _terminate_process(self):
        if self.process is not None:
            try:
                self.process.terminate()
                self.process.wait()
            except Exception as e:
                _LOG.error('Failed to wait for termination of kafka process', exc_info=e)
            finally:
                self.process = None

    def _wait_for_zk_absence(self):
        try:
            self.id_manager.wait_for_broker_id_absence()
        except Exception as e:
            _LOG.error('Failed to wait until broker id absence in zk', exc_info=e)

    def get_zk_connect_string(self):
        return self.kafka_properties.get_property('zookeeper.connect')

    def start_kafka_process(self, zookeeper_address):
        """
        Starts kafka using zookeeper address provided.
        :param zookeeper_address: Address to use for kafka
        :raise LeaderElectionInProgress: raised when broker can not be started because leader election is in progress
        """
        if not self.process:
            self._verify_leaders_election_progress()

            broker_id = self.id_manager.get_broker_id()
            _LOG.info('Using broker_id {} for kafka'.format(broker_id))
            if broker_id is not None:
                self.kafka_properties.set_property('broker.id', broker_id)
            else:
                self.kafka_properties.delete_property('broker.id')

            _LOG.info('Using ZK address: {}'.format(zookeeper_address))
            self.kafka_properties.set_property('zookeeper.connect', zookeeper_address)

            self.kafka_properties.dump()

            _LOG.info('Staring kafka process')
            self.process = self._open_process()

            _LOG.info('Waiting for kafka to start up with timeout {} seconds'.format(self.wait_timeout))
            if not self.id_manager.wait_for_broker_id_presence(self.wait_timeout):
                self.wait_timeout += 60
                _LOG.error(
                    'Failed to wait for broker to start up, probably will kill, increasing timeout to {} seconds'.format(
                        self.wait_timeout))

    def _verify_leaders_election_progress(self):
        if self._is_clean_election():
            active_brokers = self.exhibitor.get_children('/brokers/ids')

            for topic in self.exhibitor.get_children('/brokers/topics'):
                for partition in self.exhibitor.get_children('/brokers/topics/{}/partitions'.format(topic)):
                    state = json.loads(
                        self.exhibitor.get('/brokers/topics/{}/partitions/{}/state'.format(topic, partition))[0].decode(
                            'utf-8'))
                    if str(state['leader']) in active_brokers:
                        _LOG.warn('Leadership is not transferred for {} {} ({}, brokers: {})'.format(topic, partition,
                                                                                                     json.dumps(state),
                                                                                                     active_brokers))
                        raise LeaderElectionInProgress()
                    if any([str(x) in active_brokers for x in state['isr']]):
                        _LOG.warn('Leadership is not transferred for {} {} ({}, brokers: {})'.format(topic, partition,
                                                                                                     json.dumps(state),
                                                                                                     active_brokers))
                        raise LeaderElectionInProgress()

    def _open_process(self):
        return subprocess.Popen(
            [self.kafka_dir + "/bin/kafka-server-start.sh", self.kafka_properties.settings_file])



import logging
import random
import time

import requests
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError
from requests.exceptions import RequestException

_LOG = logging.getLogger('bubuku.exhibitor')


class ExhibitorEnsembleProvider:
    TIMEOUT = 3.1

    def __init__(self, hosts, port, uri_path='/exhibitor/v1/cluster/list', poll_interval=300):
        self._exhibitor_port = port
        self._uri_path = uri_path
        self._poll_interval = poll_interval
        self._exhibitors = hosts
        self._master_exhibitors = hosts
        self._zookeeper_hosts = ''
        self._next_poll = None
        while not self.poll():
            _LOG.info('waiting on exhibitor')
            time.sleep(5)

    def poll(self):
        if self._next_poll and self._next_poll > time.time():
            return False

        json = self._query_exhibitors(self._exhibitors)
        if not json:
            json = self._query_exhibitors(self._master_exhibitors)

        if isinstance(json, dict) and 'servers' in json and 'port' in json:
            self._next_poll = time.time() + self._poll_interval
            zookeeper_hosts = ','.join([h + ':' + str(json['port']) for h in sorted(json['servers'])])
            if self._zookeeper_hosts != zookeeper_hosts:
                _LOG.info('ZooKeeper connection string has changed: %s => %s', self._zookeeper_hosts, zookeeper_hosts)
                self._zookeeper_hosts = zookeeper_hosts
                self._exhibitors = json['servers']
                return True
        return False

    def _query_exhibitors(self, exhibitors):
        if exhibitors == [None]:
            return {'servers': ['localhost'], 'port': 2181}
        random.shuffle(exhibitors)
        for host in exhibitors:
            uri = 'http://{}:{}{}'.format(host, self._exhibitor_port, self._uri_path)
            try:
                response = requests.get(uri, timeout=self.TIMEOUT)
                return response.json()
            except RequestException:
                pass
        return None

    @property
    def zookeeper_hosts(self):
        return self._zookeeper_hosts


class Exhibitor:
    def __init__(self, hosts, port, prefix):
        self.prefix = prefix
        self.exhibitor = ExhibitorEnsembleProvider(hosts, port, poll_interval=30)
        self.client = KazooClient(hosts=self.exhibitor.zookeeper_hosts + self.prefix,
                                  command_retry={
                                      'deadline': 10,
                                      'max_delay': 1,
                                      'max_tries': -1},
                                  connection_retry={'max_delay': 1, 'max_tries': -1})
        self.client.add_listener(self.session_listener)
        self.client.start()

    def session_listener(self, state):
        pass

    def _poll_exhibitor(self):
        if self.exhibitor.poll():
            self.client.stop()
            self.client.set_hosts(self.exhibitor.zookeeper_hosts + self.prefix)
            self.client.start()

    def get(self, *params):
        self._poll_exhibitor()
        return self.client.retry(self.client.get, *params)

    def set(self, *args, **kwargs):
        self._poll_exhibitor()
        return self.client.retry(self.client.set, *args, **kwargs)

    def create(self, *args, **kwargs):
        self._poll_exhibitor()
        return self.client.retry(self.client.create, *args, **kwargs)

    def delete(self, *args, **kwargs):
        self._poll_exhibitor()
        return self.client.retry(self.client.delete, *args, **kwargs)

    def get_children(self, *params):
        self._poll_exhibitor()
        try:
            return self.client.retry(self.client.get_children, *params)
        except NoNodeError:
            return []

    def take_lock(self, *args, **kwargs):
        while True:
            try:
                self._poll_exhibitor()
                return self.client.Lock(*args, **kwargs)
            except Exception as e:
                _LOG.error('Failed to obtain lock for exhibitor, retrying', exc_info=e)


def load_exhibitor(initial_hosts: list, zookeeper_prefix):
    return Exhibitor(initial_hosts, 8181, zookeeper_prefix)



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Change, Check
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.features.restart_on_zk')


class RestartBrokerChange(Change):
    def __init__(self, zk_hosts: str, zk: Exhibitor, broker: BrokerManager):
        self.zk_hosts = zk_hosts
        self.zk = zk
        self.broker = broker

    def get_name(self):
        return 'restart'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['start', 'restart', 'stop']])

    def run(self, current_actions):
        current_hosts = self.zk.exhibitor.zookeeper_hosts
        if current_hosts != self.zk_hosts:
            _LOG.warning('ZK address changed again, from {} to {}'.format(self.zk_hosts, current_hosts))
            return False

        new_zk_connect = self.zk_hosts + self.zk.prefix
        if self.broker.get_zk_connect_string() != new_zk_connect:
            _LOG.info('Executing restart to update ZK to {}'.format(new_zk_connect))
            self.broker.stop_kafka_process()
            if self.broker.has_leadership():
                return True
        else:
            _LOG.info('ZK connect string already configured correctly {}, will not restart'.format(new_zk_connect))
        return False


class CheckExhibitorAddressChanged(Check):
    def __init__(self, zk: Exhibitor, broker: BrokerManager):
        self.zk = zk
        self.broker = broker
        self.zk_address = None

    def check(self) -> Change:
        new_addresses = self.zk.exhibitor.zookeeper_hosts
        if new_addresses != self.zk_address:
            _LOG.info('ZK addresses changed from {} to {}, triggering restart'.format(self.zk_address, new_addresses))
            self.zk_address = new_addresses
            return RestartBrokerChange(new_addresses, self.zk, self.broker)



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Change, Check
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.features.restart_if_dead')


class StartBrokerChange(Change):
    def __init__(self, broker: BrokerManager, zk: Exhibitor):
        self.broker = broker
        self.zk = zk
        self.stopped = False

    def get_name(self) -> str:
        return 'start'

    def can_run(self, current_actions):
        # Only one restart allowed.
        return all([action not in current_actions for action in ['start', 'restart', 'stop']])

    def run(self, current_actions) -> bool:
        if self.broker.is_running_and_registered():
            return False
        _LOG.info('Waiting for complete death')
        if not self.stopped:
            self.broker.stop_kafka_process()
            self.stopped = True

        if self.broker.has_leadership():
            return True  # Wait for leadership transfer
        _LOG.info('Starting up again')
        try:
            self.broker.start_kafka_process(self.zk.exhibitor.zookeeper_hosts + self.zk.prefix)
        except Exception as ex:
            _LOG.warn('Failed to start kafka process', exc_info=ex)
            return True
        return False


class CheckBrokerStopped(Check):
    def __init__(self, broker: BrokerManager, zk: Exhibitor):
        self.broker = broker
        self.zk = zk

    def check(self) -> Change:
        if self.broker.is_running_and_registered():
            return None
        _LOG.info('Oops! Broker is dead, triggering restart')
        return StartBrokerChange(self.broker, self.zk)



#!/usr/bin/env python3
import functools
import logging
import os
import re
from time import sleep, time

from kazoo.client import NoNodeError

from bubuku.amazon import Amazon
from bubuku.config import KafkaProperties
from bubuku.zookeeper import Exhibitor

_LOG = logging.getLogger('bubuku.id_generator')


class BrokerIdGenerator(object):
    def __init__(self, zk: Exhibitor):
        self.zk = zk

    def get_broker_id(self) -> str:
        raise NotImplementedError('Not implemented')

    def wait_for_broker_id_absence(self):
        while self.is_registered():
            sleep(1)

    def _is_registered_in_zk(self, id_):
        try:
            _, stat = self.zk.get('/brokers/ids/{}'.format(id_))
            return stat is not None
        except NoNodeError:
            return False

    def wait_for_broker_id_presence(self, timeout) -> bool:
        start = time()
        while not self.is_registered():
            if (time() - start) > timeout:
                return False
            sleep(1)
        return True

    def is_registered(self):
        raise NotImplementedError('Not implemented')


def _create_rfc1918_address_hash(ip: str) -> (str, str):
    address = [int(v) for v in ip.split('.')]
    # the goal of this hashing is to get positive 4-bytes int which can not be changed during restarts
    if address[0] == 10:
        address[0] = 1
    elif address[0] == 192 and address[1] == 168:
        address[0] = 2
    elif address[0] == 172 and (address[1] & 0xF0) == 16:
        address[0] = 3
    else:
        return None
    return str(functools.reduce(lambda o, v: o * 256 + v, address, 0)), str(256 * 256 * 256 * 3 + 1)


class BrokerIDByIp(BrokerIdGenerator):
    def __init__(self, zk: Exhibitor, ip: str, kafka_props: KafkaProperties):
        super().__init__(zk)
        self.broker_id, max_id = _create_rfc1918_address_hash(ip)
        kafka_props.set_property('reserved.broker.max.id', max_id)
        _LOG.info('Built broker id {} from ip: {}'.format(self.broker_id, ip))
        if self.broker_id is None:
            raise NotImplementedError('Broker id from ip address supported only for rfc1918 private addresses')

    def get_broker_id(self):
        return self.broker_id

    def is_registered(self):
        return self._is_registered_in_zk(self.broker_id)


class BrokerIdAutoAssign(BrokerIdGenerator):
    def __init__(self, zk: Exhibitor, kafka_properties: KafkaProperties):
        super().__init__(zk)
        self.kafka_properties = kafka_properties
        self.broker_id = None

    def get_broker_id(self):
        return None

    def is_registered(self):
        meta_path = '{}/meta.properties'.format(self.kafka_properties.get_property('log.dirs'))
        while not os.path.isfile(meta_path):
            return False
        with open(meta_path) as f:
            lines = f.readlines()
            for line in lines:
                match = re.search('broker\.id=(\d+)', line)
                if match:
                    return self._is_registered_in_zk(match.group(1))
        return False


def get_broker_id_policy(policy: str, zk: Exhibitor, kafka_props: KafkaProperties, amazon: Amazon) -> BrokerIdGenerator:
    if policy == 'ip':
        return BrokerIDByIp(zk, amazon.get_own_ip(), kafka_props)
    elif policy == 'auto':
        return BrokerIdAutoAssign(zk, kafka_props)
    else:
        raise Exception('Unsupported id generator policy')
